package org.eclipse.jdt.internal.compiler.flow;

import org.eclipse.jdt.internal.compiler.ast.IfStatement;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;

public abstract class FlowInfo
{
    public int tagBits;
    public static final int REACHABLE = 0;
    public static final int UNREACHABLE_OR_DEAD = 1;
    public static final int UNREACHABLE_BY_NULLANALYSIS = 2;
    public static final int UNREACHABLE = 3;
    public static final int NULL_FLAG_MASK = 4;
    public static final int UNKNOWN = 1;
    public static final int NULL = 2;
    public static final int NON_NULL = 4;
    public static final int POTENTIALLY_UNKNOWN = 8;
    public static final int POTENTIALLY_NULL = 16;
    public static final int POTENTIALLY_NON_NULL = 32;
    public static final int UNROOTED = 64;
    public static final int FREE_TYPEVARIABLE = 48;
    public static final UnconditionalFlowInfo DEAD_END;
    
    static {
        DEAD_END = new UnconditionalFlowInfo();
        FlowInfo.DEAD_END.tagBits = 3;
    }
    
    public FlowInfo() {
    }
    
    public abstract FlowInfo addInitializationsFrom(FlowInfo p0);
    
    public abstract FlowInfo addNullInfoFrom(FlowInfo p0);
    
    public abstract FlowInfo addPotentialInitializationsFrom(FlowInfo p0);
    
    public FlowInfo asNegatedCondition() {
        return this;
    }
    
    public static FlowInfo conditional(FlowInfo initsWhenTrue, FlowInfo initsWhenFalse) {
        if (initsWhenTrue == initsWhenFalse) {
            return initsWhenTrue;
        }
        return new ConditionalFlowInfo(initsWhenTrue, initsWhenFalse);
    }
    
    public boolean cannotBeDefinitelyNullOrNonNull(LocalVariableBinding local) {
        return this.isPotentiallyUnknown(local) || (this.isPotentiallyNonNull(local) && this.isPotentiallyNull(local));
    }
    
    public boolean cannotBeNull(LocalVariableBinding local) {
        return this.isDefinitelyNonNull(local) || this.isProtectedNonNull(local);
    }
    
    public boolean canOnlyBeNull(LocalVariableBinding local) {
        return this.isDefinitelyNull(local) || this.isProtectedNull(local);
    }
    
    public abstract FlowInfo copy();
    
    public static UnconditionalFlowInfo initial(int maxFieldCount) {
        UnconditionalFlowInfo info = new UnconditionalFlowInfo();
        info.maxFieldCount = maxFieldCount;
        return info;
    }
    
    public abstract FlowInfo initsWhenFalse();
    
    public abstract FlowInfo initsWhenTrue();
    
    public abstract boolean isDefinitelyAssigned(FieldBinding p0);
    
    public abstract boolean isDefinitelyAssigned(LocalVariableBinding p0);
    
    public abstract boolean isDefinitelyNonNull(LocalVariableBinding p0);
    
    public abstract boolean isDefinitelyNull(LocalVariableBinding p0);
    
    public abstract boolean isDefinitelyUnknown(LocalVariableBinding p0);
    
    public abstract boolean hasNullInfoFor(LocalVariableBinding p0);
    
    public abstract boolean isPotentiallyAssigned(FieldBinding p0);
    
    public abstract boolean isPotentiallyAssigned(LocalVariableBinding p0);
    
    public abstract boolean isPotentiallyNonNull(LocalVariableBinding p0);
    
    public abstract boolean isPotentiallyNull(LocalVariableBinding p0);
    
    public abstract boolean isPotentiallyUnknown(LocalVariableBinding p0);
    
    public abstract boolean isProtectedNonNull(LocalVariableBinding p0);
    
    public abstract boolean isProtectedNull(LocalVariableBinding p0);
    
    public abstract void markAsComparedEqualToNonNull(LocalVariableBinding p0);
    
    public abstract void markAsComparedEqualToNull(LocalVariableBinding p0);
    
    public abstract void markAsDefinitelyAssigned(FieldBinding p0);
    
    public abstract void markAsDefinitelyNonNull(LocalVariableBinding p0);
    
    public abstract void markAsDefinitelyNull(LocalVariableBinding p0);
    
    public abstract void resetNullInfo(LocalVariableBinding p0);
    
    public abstract void markPotentiallyUnknownBit(LocalVariableBinding p0);
    
    public abstract void markPotentiallyNullBit(LocalVariableBinding p0);
    
    public abstract void markPotentiallyNonNullBit(LocalVariableBinding p0);
    
    public abstract void markAsDefinitelyAssigned(LocalVariableBinding p0);
    
    public abstract void markAsDefinitelyUnknown(LocalVariableBinding p0);
    
    public void markNullStatus(LocalVariableBinding local, int nullStatus) {
        switch (nullStatus) {
            case 1: {
                this.markAsDefinitelyUnknown(local);
                break;
            }
            case 2: {
                this.markAsDefinitelyNull(local);
                break;
            }
            case 4: {
                this.markAsDefinitelyNonNull(local);
                break;
            }
            default: {
                this.resetNullInfo(local);
                if ((nullStatus & 0x8) != 0x0) {
                    this.markPotentiallyUnknownBit(local);
                }
                if ((nullStatus & 0x10) != 0x0) {
                    this.markPotentiallyNullBit(local);
                }
                if ((nullStatus & 0x20) != 0x0) {
                    this.markPotentiallyNonNullBit(local);
                }
                if ((nullStatus & 0x38) == 0x0) {
                    this.markAsDefinitelyUnknown(local);
                    break;
                }
                break;
            }
        }
    }
    
    public int nullStatus(LocalVariableBinding local) {
        if (this.isDefinitelyUnknown(local)) {
            return 1;
        }
        if (this.isDefinitelyNull(local)) {
            return 2;
        }
        if (this.isDefinitelyNonNull(local)) {
            return 4;
        }
        int status = 0;
        if (this.isPotentiallyUnknown(local)) {
            status |= 0x8;
        }
        if (this.isPotentiallyNull(local)) {
            status |= 0x10;
        }
        if (this.isPotentiallyNonNull(local)) {
            status |= 0x20;
        }
        if (status > 0) {
            return status;
        }
        return 1;
    }
    
    public static int mergeNullStatus(int nullStatus1, int nullStatus2) {
        boolean canBeNull = false;
        boolean canBeNonNull = false;
        switch (nullStatus1) {
            case 16: {
                canBeNonNull = true;
            }
            case 2: {
                canBeNull = true;
                break;
            }
            case 32: {
                canBeNull = true;
            }
            case 4: {
                canBeNonNull = true;
                break;
            }
        }
        switch (nullStatus2) {
            case 16: {
                canBeNonNull = true;
            }
            case 2: {
                canBeNull = true;
                break;
            }
            case 32: {
                canBeNull = true;
            }
            case 4: {
                canBeNonNull = true;
                break;
            }
        }
        if (canBeNull) {
            if (canBeNonNull) {
                return 16;
            }
            return 2;
        }
        else {
            if (canBeNonNull) {
                return 4;
            }
            return 1;
        }
    }
    
    public static UnconditionalFlowInfo mergedOptimizedBranches(FlowInfo initsWhenTrue, boolean isOptimizedTrue, FlowInfo initsWhenFalse, boolean isOptimizedFalse, boolean allowFakeDeadBranch) {
        UnconditionalFlowInfo mergedInfo;
        if (isOptimizedTrue) {
            if (initsWhenTrue == FlowInfo.DEAD_END && allowFakeDeadBranch) {
                mergedInfo = initsWhenFalse.setReachMode(1).unconditionalInits();
            }
            else {
                mergedInfo = initsWhenTrue.addPotentialInitializationsFrom(initsWhenFalse.nullInfoLessUnconditionalCopy()).unconditionalInits();
            }
        }
        else if (isOptimizedFalse) {
            if (initsWhenFalse == FlowInfo.DEAD_END && allowFakeDeadBranch) {
                mergedInfo = initsWhenTrue.setReachMode(1).unconditionalInits();
            }
            else {
                mergedInfo = initsWhenFalse.addPotentialInitializationsFrom(initsWhenTrue.nullInfoLessUnconditionalCopy()).unconditionalInits();
            }
        }
        else {
            mergedInfo = initsWhenTrue.mergedWith(initsWhenFalse.unconditionalInits());
        }
        return mergedInfo;
    }
    
    public static UnconditionalFlowInfo mergedOptimizedBranchesIfElse(FlowInfo initsWhenTrue, boolean isOptimizedTrue, FlowInfo initsWhenFalse, boolean isOptimizedFalse, boolean allowFakeDeadBranch, FlowInfo flowInfo, IfStatement ifStatement, boolean reportDeadCodeInKnownPattern) {
        UnconditionalFlowInfo mergedInfo;
        if (isOptimizedTrue) {
            if (initsWhenTrue == FlowInfo.DEAD_END && allowFakeDeadBranch) {
                if (!reportDeadCodeInKnownPattern) {
                    if (ifStatement.elseStatement == null) {
                        mergedInfo = flowInfo.unconditionalInits();
                    }
                    else {
                        mergedInfo = initsWhenFalse.unconditionalInits();
                        if (initsWhenFalse != FlowInfo.DEAD_END) {
                            mergedInfo.setReachMode(flowInfo.reachMode());
                        }
                    }
                }
                else {
                    mergedInfo = initsWhenFalse.setReachMode(1).unconditionalInits();
                }
            }
            else {
                mergedInfo = initsWhenTrue.addPotentialInitializationsFrom(initsWhenFalse.nullInfoLessUnconditionalCopy()).unconditionalInits();
            }
        }
        else if (isOptimizedFalse) {
            if (initsWhenFalse == FlowInfo.DEAD_END && allowFakeDeadBranch) {
                if (!reportDeadCodeInKnownPattern) {
                    if (ifStatement.thenStatement == null) {
                        mergedInfo = flowInfo.unconditionalInits();
                    }
                    else {
                        mergedInfo = initsWhenTrue.unconditionalInits();
                        if (initsWhenTrue != FlowInfo.DEAD_END) {
                            mergedInfo.setReachMode(flowInfo.reachMode());
                        }
                    }
                }
                else {
                    mergedInfo = initsWhenTrue.setReachMode(1).unconditionalInits();
                }
            }
            else {
                mergedInfo = initsWhenFalse.addPotentialInitializationsFrom(initsWhenTrue.nullInfoLessUnconditionalCopy()).unconditionalInits();
            }
        }
        else if ((flowInfo.tagBits & 0x3) == 0x0 && (ifStatement.bits & 0x80) != 0x0 && initsWhenTrue != FlowInfo.DEAD_END && initsWhenFalse != FlowInfo.DEAD_END) {
            mergedInfo = initsWhenTrue.addPotentialInitializationsFrom(initsWhenFalse.nullInfoLessUnconditionalCopy()).unconditionalInits();
            mergedInfo.mergeDefiniteInitsWith(initsWhenFalse.unconditionalCopy());
            if ((mergedInfo.tagBits & 0x1) != 0x0 && (initsWhenFalse.tagBits & 0x3) == 0x2) {
                mergedInfo.tagBits &= 0xFFFFFFFE;
                mergedInfo.tagBits |= 0x2;
            }
        }
        else if ((flowInfo.tagBits & 0x3) == 0x0 && (ifStatement.bits & 0x100) != 0x0 && initsWhenTrue != FlowInfo.DEAD_END && initsWhenFalse != FlowInfo.DEAD_END) {
            mergedInfo = initsWhenFalse.addPotentialInitializationsFrom(initsWhenTrue.nullInfoLessUnconditionalCopy()).unconditionalInits();
            mergedInfo.mergeDefiniteInitsWith(initsWhenTrue.unconditionalCopy());
            if ((mergedInfo.tagBits & 0x1) != 0x0 && (initsWhenTrue.tagBits & 0x3) == 0x2) {
                mergedInfo.tagBits &= 0xFFFFFFFE;
                mergedInfo.tagBits |= 0x2;
            }
        }
        else {
            mergedInfo = initsWhenTrue.mergedWith(initsWhenFalse.unconditionalInits());
        }
        return mergedInfo;
    }
    
    public int reachMode() {
        return this.tagBits & 0x3;
    }
    
    public abstract FlowInfo safeInitsWhenTrue();
    
    public abstract FlowInfo setReachMode(int p0);
    
    public abstract UnconditionalFlowInfo mergedWith(UnconditionalFlowInfo p0);
    
    public abstract UnconditionalFlowInfo mergeDefiniteInitsWith(UnconditionalFlowInfo p0);
    
    public abstract UnconditionalFlowInfo nullInfoLessUnconditionalCopy();
    
    @Override
    public String toString() {
        if (this == FlowInfo.DEAD_END) {
            return "FlowInfo.DEAD_END";
        }
        return super.toString();
    }
    
    public abstract UnconditionalFlowInfo unconditionalCopy();
    
    public abstract UnconditionalFlowInfo unconditionalFieldLessCopy();
    
    public abstract UnconditionalFlowInfo unconditionalInits();
    
    public abstract UnconditionalFlowInfo unconditionalInitsWithoutSideEffect();
    
    public abstract void resetAssignmentInfo(LocalVariableBinding p0);
    
    public static int tagBitsToNullStatus(long tagBits) {
        if ((tagBits & 0x100000000000000L) != 0x0L) {
            return 4;
        }
        if ((tagBits & 0x80000000000000L) != 0x0L) {
            return 48;
        }
        return 1;
    }
}
